"""
=============
Lookup Tables
=============

Simulations tend to require a large quantity of data to run.  :mod:`vivarium`
provides the :class:`LookupTable` abstraction to ensure that accurate data can
be retrieved when it's needed. It's a callable object that takes in a
population index and returns data specific to the individuals represented by
that index. See the :ref:`lookup concept note <lookup_concept>` for more.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from datetime import datetime
from typing import Any

import numpy as np
import pandas as pd

from vivarium.framework.lookup.interpolation import Interpolation
from vivarium.framework.population.population_view import PopulationView
from vivarium.types import ClockTime, ScalarValue


class LookupTable(ABC):
    """A callable to produces values for a population index.

    In :mod:`vivarium` simulations, the index is synonymous with the simulated
    population.  The lookup system allows the user to provide different kinds
    of data and strategies for using that data.  When the simulation is
    running, then, components can lookup parameter values based solely on
    the population index.

    Notes
    -----
    These should not be created directly. Use the `lookup` method on the builder
    during setup.

    """

    def __init__(
        self,
        table_number: int,
        key_columns: Sequence[str] = (),
        parameter_columns: Sequence[str] = (),
        value_columns: Sequence[str] = (),
        validate: bool = True,
    ):
        self.table_number = table_number
        """Unique identifier of the table."""
        self.key_columns = key_columns
        """Column names to be used as categorical parameters in Interpolation
        to select between interpolation functions."""
        self.parameter_columns = parameter_columns
        """Column names to be used as continuous parameters in Interpolation."""
        self.value_columns = list(value_columns)
        """Names of value columns to be interpolated over."""
        self.validate = validate
        """Whether to validate the data before building the LookupTable."""

    @property
    def name(self) -> str:
        """Tables are generically named after the order they were created."""
        return f"lookup_table_{self.table_number}"

    def __call__(self, index: pd.Index[int]) -> pd.Series[Any] | pd.DataFrame:
        """Get the mapped values for the given index.

        Parameters
        ----------
        index
            Index for population view.

        Returns
        -------
            pandas.Series if only one value_column, pandas.DataFrame if multiple
            columns

        """
        mapped_values: pd.Series[Any] | pd.DataFrame = self.call(index).squeeze(axis=1)
        return mapped_values

    @abstractmethod
    def call(self, index: pd.Index[int]) -> pd.DataFrame:
        """Private method to allow LookupManager to add constraints."""
        pass

    def __repr__(self) -> str:
        return "LookupTable()"


class InterpolatedTable(LookupTable):
    """A callable that interpolates data according to a given strategy.

    Notes
    -----
    These should not be created directly. Use the `lookup` interface on the
    :class:`builder <vivarium.framework.engine.Builder>` during setup.

    """

    def __init__(
        self,
        table_number: int,
        data: pd.DataFrame,
        population_view_builder: Callable[[list[str]], PopulationView],
        key_columns: Sequence[str],
        parameter_columns: Sequence[str],
        value_columns: Sequence[str],
        interpolation_order: int,
        clock: Callable[[], ClockTime],
        extrapolate: bool,
        validate: bool,
    ):
        super().__init__(
            table_number=table_number,
            key_columns=key_columns,
            parameter_columns=parameter_columns,
            value_columns=value_columns,
            validate=validate,
        )
        self.data = data
        self.clock = clock
        self.interpolation_order = interpolation_order
        self.extrapolate = extrapolate
        """Callable for current time in simulation."""
        param_cols_with_edges = []
        for p in parameter_columns:
            param_cols_with_edges += [(p, f"{p}_start", f"{p}_end")]
        view_columns = sorted((set(key_columns) | set(parameter_columns)) - {"year"}) + [
            "tracked"
        ]

        self.parameter_columns_with_edges = param_cols_with_edges

        required_cols = (
            set(self.key_columns)
            | set([col for p in self.parameter_columns_with_edges for col in p])
            | set(self.value_columns)
        )
        extra_columns = self.data.columns.difference(list(required_cols))

        if not self.value_columns:
            self.value_columns = list(extra_columns)
        else:
            self.data = self.data.drop(columns=extra_columns)

        self.population_view = population_view_builder(view_columns)
        self.interpolation = Interpolation(
            data,
            self.key_columns,
            self.parameter_columns_with_edges,
            self.value_columns,
            order=self.interpolation_order,
            extrapolate=self.extrapolate,
            validate=self.validate,
        )

    def call(self, index: pd.Index[int]) -> pd.DataFrame:
        """Get the interpolated values for the rows in ``index``.

        Parameters
        ----------
        index
            Index of the population to interpolate for.

        Returns
        -------
            A table with the interpolated values for the population requested.

        """
        pop = self.population_view.get(index)
        del pop["tracked"]
        if "year" in [col for col in self.parameter_columns]:
            current_time = self.clock()
            # TODO: [MIC-5478] handle Number output from clock
            if isinstance(current_time, pd.Timestamp) or isinstance(current_time, datetime):
                fractional_year = float(current_time.year)
                fractional_year += current_time.timetuple().tm_yday / 365.25
                pop["year"] = fractional_year
            else:
                raise ValueError(
                    "You cannot use the column 'year' in a simulation unless your simulation uses a DateTimeClock."
                )

        return self.interpolation(pop)

    def __repr__(self) -> str:
        return "InterpolatedTable()"


class CategoricalTable(LookupTable):
    """
    A callable that selects values from a table based on categorical parameters
    across an index.

    Notes
    -----
    These should not be created directly. Use the `lookup` interface on the
    :class:`builder <vivarium.framework.engine.Builder>` during setup.

    """

    def __init__(
        self,
        table_number: int,
        data: pd.DataFrame,
        population_view_builder: Callable[[list[str]], PopulationView],
        key_columns: Sequence[str],
        value_columns: Sequence[str],
    ):
        super().__init__(
            table_number=table_number,
            key_columns=key_columns,
            value_columns=value_columns,
        )
        self.data = data
        self.population_view = population_view_builder(list(self.key_columns) + ["tracked"])

        extra_columns = self.data.columns.difference(
            list(set(self.key_columns) | set(self.value_columns))
        )

        if not self.value_columns:
            self.value_columns = list(extra_columns)
        else:
            self.data = self.data.drop(columns=extra_columns)

    def call(self, index: pd.Index[int]) -> pd.DataFrame:
        """Get the mapped values for the rows in ``index``.

        Parameters
        ----------
        index
            Index of the population to interpolate for.

        Returns
        -------
            A table with the mapped values for the population requested.
        """
        pop = self.population_view.get(index)
        del pop["tracked"]

        # specify some numeric type for columns, so they won't be objects but
        # will be updated with whatever column type it actually is
        result = pd.DataFrame(index=pop.index, columns=self.value_columns, dtype=np.float64)

        sub_tables = pop.groupby(list(self.key_columns))
        for key, sub_table in list(sub_tables):
            if sub_table.empty:
                continue

            category_masks: list[pd.Series[bool]] = [
                self.data[self.key_columns[i]] == category for i, category in enumerate(key)
            ]
            joint_mask = pd.Series(True, index=self.data.index)
            for category_mask in category_masks:
                joint_mask = joint_mask & category_mask
            values = self.data.loc[joint_mask, self.value_columns].values
            result.loc[sub_table.index, self.value_columns] = values

        return result

    def __repr__(self) -> str:
        return "CategoricalTable()"


class ScalarTable(LookupTable):
    """A callable that broadcasts a scalar or list of scalars over an index.

    Notes
    -----
    These should not be created directly. Use the `lookup` interface on the
    builder during setup.
    """

    def __init__(
        self,
        table_number: int,
        data: ScalarValue | list[ScalarValue] | tuple[ScalarValue],
        key_columns: Sequence[str] = (),
        parameter_columns: Sequence[str] = (),
        value_columns: Sequence[str] = (),
        validate: bool = True,
    ):
        super().__init__(
            table_number, key_columns, parameter_columns, value_columns, validate
        )
        self.data = data

    def call(self, index: pd.Index[int]) -> pd.DataFrame:
        """Broadcast this tables values over the provided index.

        Parameters
        ----------
        index
            Index of the population to construct table for.

        Returns
        -------
            A table with a column for each of the scalar values for the
            population requested.

        """
        if not isinstance(self.data, (list, tuple)):
            values_series: pd.Series[Any] = pd.Series(
                self.data,
                index=index,
                name=self.value_columns[0] if self.value_columns else None,
            )
            return pd.DataFrame(values_series)
        else:
            values_list: list[pd.Series[Any]] = [pd.Series(v, index=index) for v in self.data]
            return pd.DataFrame(dict(zip(self.value_columns, values_list)))

    def __repr__(self) -> str:
        return "ScalarTable(value(s)={})".format(self.data)



from datetime import datetime, timedelta

import numpy as np
from layered_config_tree import LayeredConfigTree
from vivarium_testing_utils import FuzzyChecker

from vivarium import InteractiveContext
from vivarium.framework.utilities import from_yearly


def test_disease_model(fuzzy_checker: FuzzyChecker, disease_model_spec) -> None:
    config = LayeredConfigTree(disease_model_spec, layers=["base", "override"])
    config.update(
        {
            "configuration": {
                "mortality": {
                    "mortality_rate": 20.0,
                },
                "lower_respiratory_infections": {
                    "incidence_rate": 25.0,
                    "remission_rate": 50.0,
                    "excess_mortality_rate": 0.01,
                },
            },
        }
    )

    simulation = InteractiveContext(config)

    pop = simulation.get_population()
    expected_columns = {
        "tracked",
        "alive",
        "age",
        "sex",
        "entrance_time",
        "lower_respiratory_infections",
        "child_wasting_propensity",
    }
    assert set(pop.columns) == expected_columns
    assert len(pop) == 100_000
    assert np.all(pop["tracked"] == True)
    assert np.all(pop["alive"] == "alive")
    assert np.all((pop["age"] >= 0) & (pop["age"] <= 5))
    assert np.all(pop["entrance_time"] == datetime(2021, 12, 31, 12))

    for sex in ["Female", "Male"]:
        fuzzy_checker.fuzzy_assert_proportion(
            observed_numerator=(pop["sex"] == sex).sum(),
            observed_denominator=len(pop),
            target_proportion=0.5,
            # todo: remove this parameter when MIC-5412 is resolved
            name=f"{sex}_proportion",
        )

    assert np.all(pop["lower_respiratory_infections"] == "susceptible_to_lower_respiratory_infections")
    assert np.all((pop["child_wasting_propensity"] >= 0) & (pop["child_wasting_propensity"] <= 1))

    simulation.step()
    pop = simulation.get_population()
    is_alive = pop["alive"] == "alive"

    fuzzy_checker.fuzzy_assert_proportion(
        observed_numerator=(len(pop[~is_alive])),
        observed_denominator=len(pop),
        target_proportion=from_yearly(20, timedelta(days=0.5)),
        # todo: remove this parameter when MIC-5412 is resolved
        name="alive_proportion",
    )

    has_lri = pop["lower_respiratory_infections"] == "infected_with_lower_respiratory_infections"
    fuzzy_checker.fuzzy_assert_proportion(
        observed_numerator=(len(pop[is_alive & has_lri])),
        observed_denominator=len(pop[is_alive]),
        target_proportion=from_yearly(25, timedelta(days=0.5)),
        # todo: remove this parameter when MIC-5412 is resolved
        name="lri_proportion",
    )

    # todo test remission and excess mortality
    # todo test risk factor and intervention



from __future__ import annotations

from typing import Any

import pandas as pd
import pytest

from vivarium.framework.randomness import RESIDUAL_CHOICE


@pytest.fixture(params=[10**4, 10**5])
def index(request: pytest.FixtureRequest) -> pd.Index[int] | None:
    return pd.Index(range(request.param)) if request.param else None


@pytest.fixture(params=[["a", "small", "bird"]])
def choices(request: pytest.FixtureRequest) -> list[str]:
    return request.param  # type: ignore [no-any-return]


# TODO: Add 2-d weights to the tests.
@pytest.fixture(params=[None, [10, 10, 10], [0.5, 0.1, 0.4]])
def weights(request: pytest.FixtureRequest) -> None | list[int | float]:
    return request.param  # type: ignore [no-any-return]


@pytest.fixture(
    params=[
        (0.2, 0.1, RESIDUAL_CHOICE),
        (0.5, 0.6, RESIDUAL_CHOICE),
        (0.1, RESIDUAL_CHOICE, RESIDUAL_CHOICE),
    ]
)
def weights_with_residuals(request: pytest.FixtureRequest) -> tuple[Any]:
    return request.param  # type: ignore [no-any-return]



"""
============================
The Plugin Management System
============================

.. todo::
   This part will come in with the full description of the plugin system
   in the next PR. -J.C. 05/07/19

"""

from dataclasses import dataclass
from typing import TypeVar

from layered_config_tree.main import LayeredConfigTree

from vivarium.exceptions import VivariumError
from vivarium.framework.artifact.manager import ArtifactInterface, ArtifactManager
from vivarium.framework.components.manager import ComponentInterface, ComponentManager
from vivarium.framework.components.parser import ComponentConfigurationParser
from vivarium.framework.event import EventInterface, EventManager
from vivarium.framework.lifecycle import LifeCycleInterface, LifeCycleManager
from vivarium.framework.logging.manager import LoggingInterface, LoggingManager
from vivarium.framework.lookup.manager import LookupTableInterface, LookupTableManager
from vivarium.framework.population.manager import PopulationInterface, PopulationManager
from vivarium.framework.randomness.manager import RandomnessInterface, RandomnessManager
from vivarium.framework.resource.manager import ResourceInterface, ResourceManager
from vivarium.framework.results.interface import ResultsInterface
from vivarium.framework.results.manager import ResultsManager
from vivarium.framework.time import SimulationClock, TimeInterface
from vivarium.framework.utilities import import_by_path
from vivarium.framework.values.manager import ValuesInterface, ValuesManager
from vivarium.manager import Interface, Manager

I = TypeVar("I", bound=Interface)
M = TypeVar("M", bound=Manager)


_MANAGERS = {
    "logging": {
        "controller": "vivarium.framework.logging.LoggingManager",
        "builder_interface": "vivarium.framework.logging.LoggingInterface",
    },
    "lookup": {
        "controller": "vivarium.framework.lookup.LookupTableManager",
        "builder_interface": "vivarium.framework.lookup.LookupTableInterface",
    },
    "randomness": {
        "controller": "vivarium.framework.randomness.RandomnessManager",
        "builder_interface": "vivarium.framework.randomness.RandomnessInterface",
    },
    "value": {
        "controller": "vivarium.framework.values.ValuesManager",
        "builder_interface": "vivarium.framework.values.ValuesInterface",
    },
    "event": {
        "controller": "vivarium.framework.event.EventManager",
        "builder_interface": "vivarium.framework.event.EventInterface",
    },
    "population": {
        "controller": "vivarium.framework.population.PopulationManager",
        "builder_interface": "vivarium.framework.population.PopulationInterface",
    },
    "resource": {
        "controller": "vivarium.framework.resource.ResourceManager",
        "builder_interface": "vivarium.framework.resource.ResourceInterface",
    },
}
DEFAULT_PLUGINS = {
    "plugins": {
        "required": {
            "component_manager": {
                "controller": "vivarium.framework.components.ComponentManager",
                "builder_interface": "vivarium.framework.components.ComponentInterface",
            },
            "clock": {
                "controller": "vivarium.framework.time.DateTimeClock",
                "builder_interface": "vivarium.framework.time.TimeInterface",
            },
            "component_configuration_parser": {
                "controller": "vivarium.framework.components.ComponentConfigurationParser",
            },
            "lifecycle": {
                "controller": "vivarium.framework.lifecycle.LifeCycleManager",
                "builder_interface": "vivarium.framework.lifecycle.LifeCycleInterface",
            },
            "data": {
                "controller": "vivarium.framework.artifact.ArtifactManager",
                "builder_interface": "vivarium.framework.artifact.ArtifactInterface",
            },
            "results": {
                "controller": "vivarium.framework.results.ResultsManager",
                "builder_interface": "vivarium.framework.results.ResultsInterface",
            },
        },
        "optional": {},
    }
}
MANAGER_TO_STRING_MAPPER: dict[type[Manager], str] = {
    SimulationClock: "clock",
    ComponentManager: "component_manager",
    ArtifactManager: "data",
    LifeCycleManager: "lifecycle",
    ResultsManager: "results",
    LoggingManager: "logging",
    ValuesManager: "value",
    EventManager: "event",
    PopulationManager: "population",
    ResourceManager: "resource",
    LookupTableManager: "lookup",
    RandomnessManager: "randomness",
}
INTERFACE_TO_STRING_MAPPER: dict[type[Interface], str] = {
    LoggingInterface: "logging",
    LookupTableInterface: "lookup",
    ValuesInterface: "value",
    EventInterface: "event",
    TimeInterface: "clock",
    PopulationInterface: "population",
    ResourceInterface: "resource",
    ResultsInterface: "results",
    RandomnessInterface: "randomness",
    ComponentInterface: "component_manager",
    LifeCycleInterface: "lifecycle",
    ArtifactInterface: "data",
}


@dataclass
class PluginGroup:
    controller: Manager | ComponentConfigurationParser
    builder_interface: Interface


class PluginConfigurationError(VivariumError):
    """Error raised when plugin configuration is incorrectly specified."""

    pass


class PluginManager(Manager):
    @property
    def name(self) -> str:
        return "plugin_manager"

    def __init__(
        self,
        plugin_configuration: (
            dict[str, dict[str, dict[str, str]]] | LayeredConfigTree | None
        ) = None,
    ):
        self._plugin_configuration = LayeredConfigTree(
            DEFAULT_PLUGINS["plugins"], layers=["base", "override"]
        )
        self._plugin_configuration.update(plugin_configuration, source="initialization_args")
        self._plugins: dict[str, PluginGroup] = {}

    def get_plugin(self, manager_type: type[M]) -> M:
        name = self.get_manager_name(manager_type)
        manager = self.get_plugin_from_name(name)
        if not isinstance(manager, manager_type):
            raise PluginConfigurationError(
                f"Plugin {name} does not implement the correct interface."
            )
        return manager

    def get_plugin_from_name(self, name: str) -> Manager:
        if name not in self._plugins:
            self._plugins[name] = self._get(name)
        manager = self._plugins[name].controller
        if not isinstance(manager, Manager):
            raise PluginConfigurationError(
                f"Plugin {name} does not implement the correct interface."
            )
        return manager

    def get_interface_from_name(self, name: str) -> Interface:
        if name not in self._plugins:
            self._plugins[name] = self._get(name)
        interface = self._plugins[name].builder_interface
        return interface

    def get_plugin_interface(self, interface_type: type[I]) -> I:
        name = self.get_interface_name(interface_type)
        interface = self.get_interface_from_name(name)
        if not isinstance(interface, interface_type):
            raise PluginConfigurationError(
                f"Plugin {name} does not implement the correct interface."
            )
        return interface

    def get_optional_controllers(self) -> dict[str, Manager]:
        return {
            name: self.get_plugin_from_name(name)
            for name in self._plugin_configuration["optional"].keys()
        }

    def get_optional_interfaces(self) -> dict[str, Interface]:
        return {
            name: self.get_interface_from_name(name)
            for name in self._plugin_configuration["optional"].keys()
        }

    def _get(self, name: str) -> PluginGroup:
        if name not in self._plugins:
            self._plugins[name] = self._build_plugin(name)
        return self._plugins[name]

    def _build_plugin(self, name: str) -> PluginGroup:
        plugin = self._lookup(name)

        try:
            controller = import_by_path(plugin["controller"])()
        except ValueError:
            raise PluginConfigurationError(
                f'Invalid plugin specification {plugin["controller"]}'
            )

        try:
            interface_name = plugin.get("builder_interface")
            if interface_name:
                interface = import_by_path(interface_name)(controller)
            else:
                interface = None
        except ValueError:
            raise PluginConfigurationError(
                f'Invalid plugin specification {plugin["builder_interface"]}'
            )

        return PluginGroup(controller=controller, builder_interface=interface)

    def _lookup(self, name: str) -> dict[str, str]:
        if name in self._plugin_configuration["required"]:
            return self._plugin_configuration.get_tree("required").get_tree(name).to_dict()
        elif name in self._plugin_configuration["optional"]:
            return self._plugin_configuration.get_tree("optional").get_tree(name).to_dict()
        elif name in _MANAGERS:
            return _MANAGERS[name]
        else:
            raise PluginConfigurationError(f"Plugin {name} not found.")

    def __repr__(self) -> str:
        return "PluginManager()"

    def get_component_config_parser(self) -> ComponentConfigurationParser:
        name = "component_configuration_parser"
        self._plugins[name] = self._get(name)
        component_config_parser = self._plugins[name].controller
        if not isinstance(component_config_parser, ComponentConfigurationParser):
            raise PluginConfigurationError(
                f"Plugin {name} does not implement the correct interface."
            )
        return component_config_parser

    def get_manager_name(self, manager_type: type[Manager]) -> str:
        return MANAGER_TO_STRING_MAPPER[manager_type]

    def get_interface_name(self, interface_type: type[Interface]) -> str:
        return INTERFACE_TO_STRING_MAPPER[interface_type]

    def get_manager_type_from_name(self, name: str) -> type[Manager]:
        reverse_mapper = {v: k for k, v in MANAGER_TO_STRING_MAPPER.items()}
        return reverse_mapper[name]

    def get_interface_type_from_name(self, name: str) -> type[Interface]:
        reverse_mapper = {v: k for k, v in INTERFACE_TO_STRING_MAPPER.items()}
        return reverse_mapper[name]



from datetime import datetime

import pytest

from tests.helpers import ColumnCreator, ColumnRequirer
from vivarium.framework.randomness import RandomnessStream
from vivarium.framework.randomness.index_map import IndexMap
from vivarium.framework.resource.exceptions import ResourceError
from vivarium.framework.resource.group import ResourceGroup
from vivarium.framework.resource.resource import Column, NullResource, Resource
from vivarium.framework.values import Pipeline, ValueModifier, ValueSource


def test_resource_group() -> None:
    component = ColumnCreator()
    resources = [Column(str(i), component) for i in range(5)]
    r_dependencies = [
        Column("an_interesting_column", None),
        Pipeline("baz"),
        RandomnessStream("bar", lambda: datetime.now(), 1, IndexMap()),
        ValueSource(Pipeline("foo"), lambda: 1, None),
    ]

    rg = ResourceGroup(resources, r_dependencies)

    assert rg.component == component
    assert rg.type == "column"
    assert rg.names == [f"column.{i}" for i in range(5)]
    assert rg.initializer == component.on_initialize_simulants
    assert rg.dependencies == [
        "column.an_interesting_column",
        "value.baz",
        "stream.bar",
        "value_source.foo",
    ]
    assert list(rg) == rg.names


@pytest.mark.parametrize(
    "resource, has_initializer",
    [
        (Pipeline("foo"), False),
        (ValueSource(Pipeline("bar"), lambda: 1, ColumnCreator()), False),
        (ValueModifier(Pipeline("baz"), lambda: 1, ColumnCreator()), False),
        (Column("foo", ColumnCreator()), True),
        (RandomnessStream("bar", lambda: datetime.now(), 1, IndexMap()), False),
        (NullResource(0, ColumnCreator()), True),
    ],
)
def test_resource_group_is_initializer(resource: Resource, has_initializer: bool) -> None:
    rg = ResourceGroup([resource], [Column("bar", None)])
    assert rg.is_initialized == has_initializer


def test_resource_group_with_no_resources() -> None:
    with pytest.raises(ResourceError, match="must have at least one resource"):
        _ = ResourceGroup([], [Column("foo", None)])


def test_resource_group_with_multiple_components() -> None:
    resources = [
        ValueModifier(Pipeline("foo"), lambda: 1, ColumnCreator()),
        ValueSource(Pipeline("bar"), lambda: 2, ColumnRequirer()),
    ]

    with pytest.raises(ResourceError, match="resources must have the same component"):
        _ = ResourceGroup(resources, [])


def test_resource_group_with_multiple_resource_types() -> None:
    component = ColumnCreator()
    resources = [
        ValueModifier(Pipeline("foo"), lambda: 1, component),
        ValueSource(Pipeline("bar"), lambda: 2, component),
    ]

    with pytest.raises(ResourceError, match="resources must be of the same type"):
        _ = ResourceGroup(resources, [])



# mypy: ignore-errors
from vivarium.examples.disease_model.disease import (
    DiseaseModel,
    DiseaseState,
    DiseaseTransition,
    SISDiseaseModel,
)
from vivarium.examples.disease_model.intervention import TreatmentIntervention
from vivarium.examples.disease_model.mortality import Mortality
from vivarium.examples.disease_model.observer import DeathsObserver, YllsObserver
from vivarium.examples.disease_model.population import BasePopulation
from vivarium.examples.disease_model.risk import Risk, RiskEffect


def get_model_specification_path():
    from pathlib import Path

    p = Path(__file__).parent / "disease_model.yaml"
    return str(p)


def get_disease_model_simulation():
    from vivarium import InteractiveContext

    p = get_model_specification_path()
    return InteractiveContext(p)



"""
================================
The Population Management System
================================

This subpackage provides tools for managing the :term:`state table <State Table>`
in a :mod:`vivarium` simulation, which is the record of all simulants in a
simulation and their state. Its main tasks are managing the creation of new
simulants and providing the ability for components to view and update simulant
state safely during runtime.

"""

from vivarium.framework.population.exceptions import PopulationError
from vivarium.framework.population.manager import (
    PopulationInterface,
    PopulationManager,
    SimulantData,
)
from vivarium.framework.population.population_view import PopulationView



"""
=========================
The Value Pipeline System
=========================

The value pipeline system is a vital part of the :mod:`vivarium`
infrastructure. It allows for values that determine the behavior of individual
:term:`simulants <Simulant>` to be constructed across multiple
:ref:`components <components_concept>`.

For more information about when and how you should use pipelines in your
simulations, see the value system :ref:`concept note <values_concept>`.

"""
from vivarium.framework.values.combiners import ValueCombiner, list_combiner, replace_combiner
from vivarium.framework.values.exceptions import DynamicValueError
from vivarium.framework.values.manager import ValuesInterface, ValuesManager
from vivarium.framework.values.pipeline import Pipeline, ValueModifier, ValueSource
from vivarium.framework.values.post_processors import (
    PostProcessor,
    rescale_post_processor,
    union_post_processor,
)



import pytest

from tests.helpers import MockComponentA, MockManager
from vivarium.framework.components import ComponentConfigurationParser
from vivarium.framework.plugins import (
    DEFAULT_PLUGINS,
    PluginConfigurationError,
    PluginManager,
)
from vivarium.framework.time import DateTimeClock, SimulationClock, TimeInterface
from vivarium.manager import Manager

plugin_config = {"george": {"controller": "big_brother", "builder_interface": "minipax"}}


@pytest.fixture
def test_plugin_manager(model_specification):
    model_specification.plugins.optional.update(plugin_config)
    return PluginManager(model_specification.plugins)


def test_PluginManager_initializaiton(model_specification):
    model_specification.plugins.optional.update(plugin_config)
    plugin_manager = PluginManager(model_specification.plugins)

    assert (
        model_specification.plugins.to_dict()
        == plugin_manager._plugin_configuration.to_dict()
    )
    assert not plugin_manager._plugins


def test_PluginManager__lookup_fail(test_plugin_manager):
    with pytest.raises(PluginConfigurationError):
        test_plugin_manager._lookup("bananas")


def test_PluginManager__lookup(test_plugin_manager):
    for plugin in ["component_manager", "clock", "component_configuration_parser"]:
        assert (
            test_plugin_manager._lookup(plugin)
            == DEFAULT_PLUGINS["plugins"]["required"][plugin]
        )

    assert test_plugin_manager._lookup("george") == plugin_config["george"]


def test_PluginManager__get_fail(test_plugin_manager, mocker):
    import_by_path_mock = mocker.patch("vivarium.framework.plugins.import_by_path")

    def err(path):
        if path == "vivarium.framework.time.DateTimeClock":
            raise ValueError()

    import_by_path_mock.side_effect = err

    with pytest.raises(PluginConfigurationError):
        test_plugin_manager._get("clock")

    def err(path):
        if path == "vivarium.framework.time.DateTimeClock":
            return lambda: "fake_controller"
        elif path == "vivarium.framework.time.TimeInterface":
            raise ValueError()

    import_by_path_mock.side_effect = err

    with pytest.raises(PluginConfigurationError):
        test_plugin_manager._get("clock")


def test_PluginManager__get(test_plugin_manager):
    time_components = test_plugin_manager._get("clock")

    assert isinstance(time_components.controller, DateTimeClock)
    assert isinstance(time_components.builder_interface, TimeInterface)

    parser_components = test_plugin_manager._get("component_configuration_parser")

    assert isinstance(parser_components.controller, ComponentConfigurationParser)
    assert parser_components.builder_interface is None


def test_PluginManager_get_plugin(test_plugin_manager):
    assert test_plugin_manager._plugins == {}
    clock = test_plugin_manager.get_plugin(SimulationClock)
    assert isinstance(clock, DateTimeClock)
    assert test_plugin_manager._plugins["clock"].controller is clock


def test_PluginManager_get_plugin_interface(test_plugin_manager):
    assert test_plugin_manager._plugins == {}
    clock_interface = test_plugin_manager.get_plugin_interface(TimeInterface)
    assert isinstance(clock_interface, TimeInterface)
    assert test_plugin_manager._plugins["clock"].builder_interface is clock_interface


def test_PluginManager_get_optional_controllers(test_plugin_manager, mocker):
    import_by_path_mock = mocker.patch("vivarium.framework.plugins.import_by_path")
    manager = MockManager("george")

    def import_by_path_side_effect(arg):
        if arg == "big_brother":
            return lambda: manager
        else:
            return lambda _: manager

    import_by_path_mock.side_effect = import_by_path_side_effect
    assert test_plugin_manager.get_optional_controllers() == {"george": manager}
    assert import_by_path_mock.mock_calls == [
        mocker.call(plugin_config["george"]["controller"]),
        mocker.call(plugin_config["george"]["builder_interface"]),
    ]


def test_PluginManager_get_optional_interfaces(test_plugin_manager, mocker):
    import_by_path_mock = mocker.patch("vivarium.framework.plugins.import_by_path")
    component = MockComponentA("george")

    def import_by_path_side_effect(arg):
        if arg == "big_brother":
            return lambda: component
        else:
            return lambda _: component

    import_by_path_mock.side_effect = import_by_path_side_effect
    assert test_plugin_manager.get_optional_interfaces() == {"george": component}
    assert import_by_path_mock.mock_calls == [
        mocker.call(plugin_config["george"]["controller"]),
        mocker.call(plugin_config["george"]["builder_interface"]),
    ]



import math

import numpy as np
import pandas as pd
import pytest

from tests.helpers import Listener, MockComponentA, MockComponentB, MockGenericComponent
from vivarium.framework.engine import SimulationContext as SimulationContext_
from vivarium.framework.event import Event
from vivarium.framework.time import SimulationClock, get_time_stamp
from vivarium.framework.utilities import from_yearly
from vivarium.framework.values import ValuesManager, rescale_post_processor


@pytest.fixture
def builder(mocker):
    builder = mocker.MagicMock()
    manager = ValuesManager()
    manager.setup(builder)
    builder.value.register_value_producer = manager.register_value_producer
    builder.value.register_value_modifier = manager.register_value_modifier
    return builder


@pytest.fixture()
def SimulationContext():
    yield SimulationContext_
    SimulationContext_._clear_context_cache()


@pytest.fixture
def components():
    return [
        MockComponentA("gretchen", "whimsy"),
        Listener("listener"),
        MockComponentB("spoon", "antelope", "23"),
    ]


def validate_step_column_is_pipeline(sim):
    """Ensure that the pipeline and column step sizes are aligned"""
    step_pipeline = sim._values.get_value("simulant_step_size")(sim.get_population().index)
    step_column = sim._population._population.step_size
    assert np.all(step_pipeline == step_column)


def validate_index_aligned(sim, expected_active_simulants):
    """Ensure that the active simulants are as expected BEFORE a step"""
    active_simulants = sim._clock.get_active_simulants(
        get_full_pop_index(sim), sim._clock.event_time
    )

    assert active_simulants.equals(expected_active_simulants)
    assert active_simulants.difference(expected_active_simulants).empty


def validate_event_indexes(listener, expected_simulants):
    """Make sure AFTER a step, that simulants were included in the right events"""
    for index in listener.event_indexes.values():
        assert index.equals(expected_simulants)


def take_step(sim):
    old_time = sim._clock.time
    sim.step()
    new_time = sim._clock.time
    return new_time - old_time


def get_full_pop_index(sim):
    return sim.get_population().index


def get_index_by_parity(index, parity):
    if parity == "evens":
        return index[index % 2 == 0]
    elif parity == "odds":
        return index[index % 2 == 1]
    else:
        return index


def get_pop_by_parity(sim, parity):
    pop = sim.get_population()
    return pop.loc[get_index_by_parity(pop.index, parity)]


def pipeline_by_parity(sim, step_modifiers, parity):
    if parity == "all":
        return pd.concat(
            [
                pipeline_by_parity(sim, step_modifiers, "evens"),
                pipeline_by_parity(sim, step_modifiers, "odds"),
            ]
        ).sort_index()
    return pd.Series(
        from_yearly(1.75, pd.Timedelta(days=step_modifiers[parity])),
        index=get_index_by_parity(get_full_pop_index(sim), parity),
    )


def take_step_and_validate(sim, listener, expected_simulants, expected_step_size_days):
    """Take a step, and ensure that we included the right simulants, with the right step size"""
    ## Check Before Timestep
    validate_index_aligned(sim, expected_simulants)
    ## Check Timestep
    assert take_step(sim) == pd.Timedelta(days=expected_step_size_days)
    ## Check After
    validate_event_indexes(listener, expected_simulants)


class StepModifier(MockGenericComponent):
    """This mock component modifies the step size of simulants based on their index
    Odd simulants get one step size, and even simulants get another.
    """

    def __init__(
        self, name, step_modifier_even, step_modifier_odd=None, modified_simulants="all"
    ):
        super().__init__(name)
        self.step_modifier_even = step_modifier_even
        self.step_modifier_odd = (
            step_modifier_odd if step_modifier_odd else step_modifier_even
        )
        self.modified_simulants = modified_simulants

    def setup(self, builder) -> None:
        super().setup(builder)
        builder.time.register_step_size_modifier(self.modify_step)

    def modify_step(self, index):
        step_sizes = pd.Series(pd.Timedelta(days=1), index=index)
        step_sizes.loc[get_index_by_parity(index, "evens")] = pd.Timedelta(
            days=self.step_modifier_even
        )
        step_sizes.loc[get_index_by_parity(index, "odds")] = pd.Timedelta(
            days=self.step_modifier_odd
        )
        step_sizes = step_sizes.loc[get_index_by_parity(index, self.modified_simulants)]
        return step_sizes


class StepModifierWithRatePipeline(StepModifier):
    """
    Add a test pipeline that is registered, whose value is cached to
    self.ts_pipeline_value every timestep. This is meant to ensure that the
    value of a pipeline on a previous timestep was appropriately rescaled
    to the step that was actually taken.
    """

    def __init__(
        self, name, step_modifier_even, step_modifier_odd=None, modified_simulants="all"
    ):
        super().__init__(name, step_modifier_even, step_modifier_odd, modified_simulants)
        self.ts_pipeline_value = None

    def setup(self, builder) -> None:
        super().setup(builder)
        self.rate_pipeline = builder.value.register_value_producer(
            f"test_rate_{self.name}",
            source=lambda idx: pd.Series(1.75, index=idx),
            preferred_post_processor=rescale_post_processor,
        )

    def on_time_step(self, event: Event) -> None:
        self.ts_pipeline_value = self.rate_pipeline(event.index)


class StepModifierWithUntracking(StepModifierWithRatePipeline):
    """Add an event step that untracks/tracks even simulants every timestep"""

    @property
    def columns_required(self) -> list[str]:
        return ["tracked"]

    def on_time_step(self, event: Event) -> None:
        super().on_time_step(event)
        evens = self.population_view.get(event.index).loc[
            get_index_by_parity(event.index, "evens")
        ]
        evens["tracked"] = False
        self.population_view.update(evens)


class StepModifierWithMovement(StepModifierWithRatePipeline):
    def setup(self, builder) -> None:
        super().setup(builder)
        self.move_simulants_to_end = builder.time.move_simulants_to_end()

    def on_time_step(self, event: Event) -> None:
        super().on_time_step(event)
        self.move_simulants_to_end(get_index_by_parity(event.index, "evens"))


@pytest.mark.parametrize("varied_step_size", [True, False])
def test_basic_iteration(SimulationContext, base_config, components, varied_step_size):
    """Ensure that the basic iteration of the simulation works as expected.
    The step size should always be 1 in this case, the whole population should
    be updated, and the pipeline step value should always match the column step value.
    """
    if varied_step_size:
        components.append(StepModifier("step_modifier", 1))
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if hasattr(c, "args") and "listener" in c.args][0]
    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)
    assert sim._clock.time == get_time_stamp(sim.configuration.time.start)
    assert sim._clock.step_size == pd.Timedelta(days=1)
    ## Ensure that we don't have a pop view (and by extension, don't vary clocks)
    ## If no components modify the step size.
    assert bool(sim._clock._individual_clocks) == varied_step_size

    for _ in range(2):
        # After initialization, all simulants should be aligned to event times
        # After a step (and no step adjustments), simulants should still be aligned
        if varied_step_size:
            validate_step_column_is_pipeline(sim)
            assert np.all(
                sim._clock.simulant_next_event_times(full_pop_index) == sim._clock.event_time
            )
            assert np.all(
                sim._clock.simulant_step_sizes(full_pop_index) == sim._clock.step_size
            )
        take_step_and_validate(
            sim, listener, expected_simulants=full_pop_index, expected_step_size_days=1
        )


def test_empty_active_pop(SimulationContext, base_config, components):
    """Make sure that if we have no active simulants, we still take a step, given
    by the minimum step size."""
    components.append(StepModifier("step_modifier", 1))
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if hasattr(c, "args") and "listener" in c.args][0]
    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)

    ## Force a next event time update without updating step sizes.
    ## This ensures (against the current implementation) that we will have a timestep
    ## that has no simulants aligned. Check that we do the minimum timestep update.
    sim._population._population.next_event_time += pd.Timedelta(days=1)
    ## First Step
    validate_step_column_is_pipeline(sim)
    take_step_and_validate(
        sim, listener, expected_simulants=pd.Index([]), expected_step_size_days=1
    )

    ## Second Timestep
    validate_step_column_is_pipeline(sim)
    take_step_and_validate(
        sim, listener, expected_simulants=full_pop_index, expected_step_size_days=1
    )


@pytest.mark.parametrize(
    "step_modifier_even,step_modifier_odd", [(0.5, 0.5), (1, 1), (2, 2), (4.5, 4)]
)
def test_skip_iterations(
    SimulationContext, base_config, step_modifier_even, step_modifier_odd
):
    """Test that if everyone has some (non-minimum) step size, the global step adjusts to match"""
    listener = Listener("listener")
    sim = SimulationContext(
        base_config,
        [StepModifier("step_modifier", step_modifier_even, step_modifier_odd), listener],
    )
    expected_step_size = math.floor(max([step_modifier_even, 1]))
    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)

    ## Go through a couple simulant step cycles
    for _ in range(2):
        ## Everyone should update, but the step size should change
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=full_pop_index,
            expected_step_size_days=expected_step_size,
        )


def test_uneven_steps(SimulationContext, base_config):
    """Test that if we have a mix of step sizes, we take steps in accordance
    to reach all simulants' next event times in the fewest steps.
    """

    listener = Listener("listener")
    step_modifiers = {"evens": 3, "odds": 7}
    step_modifier_component = StepModifierWithRatePipeline(
        "step_modifier", step_modifiers["evens"], step_modifiers["odds"]
    )
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    ## With step sizes of 3 and 7, we need 3 steps, then 3 more, then one to get 7, then two more to get
    ## 9, etc.
    correct_step_sizes = [3, 3, 1, 2, 3, 2, 1, 3, 3]
    groups = ["evens", "evens", "odds", "evens", "evens", "odds", "evens", "evens", "all"]

    ## Ensure that steps and active simulants are correct through one cycle of 21
    for correct_step_size, group in zip(correct_step_sizes, groups):
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=get_pop_by_parity(sim, group).index,
            expected_step_size_days=correct_step_size,
        )

        sample_pipeline = step_modifier_component.ts_pipeline_value
        assert sample_pipeline.index.equals(get_pop_by_parity(sim, group).index)
        assert np.all(sample_pipeline == pipeline_by_parity(sim, step_modifiers, group))


def test_partial_modification(SimulationContext, base_config):
    """Test that if we have one modifier that doesn't apply to all simulants,
    we choose the standard value for unmodified simulants.
    """

    listener = Listener("listener")
    ## Define odds for validation, but don't pass it into the step modifier
    step_modifiers = {"evens": 3, "odds": 1}
    step_modifier_component = StepModifierWithRatePipeline(
        "step_modifier", step_modifiers["evens"], modified_simulants="evens"
    )
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    ## We should update odds with default each time, and update evens every third step.
    correct_step_sizes = [1, 1, 1, 1, 1, 1, 1, 1, 1]
    groups = ["odds", "odds", "all", "odds", "odds", "all", "odds", "odds", "all"]

    ## Ensure that steps and active simulants are correct through one cycle of 21
    for correct_step_size, group in zip(correct_step_sizes, groups):
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=get_pop_by_parity(sim, group).index,
            expected_step_size_days=correct_step_size,
        )

        sample_pipeline = step_modifier_component.ts_pipeline_value
        assert sample_pipeline.index.equals(get_pop_by_parity(sim, group).index)
        assert np.all(sample_pipeline == pipeline_by_parity(sim, step_modifiers, group))


def test_standard_step_size(SimulationContext, base_config):
    """Test that if we have one modifier that doesn't apply to all simulants,
    we choose the standard value for unmodified simulants.
    """

    base_config.update({"configuration": {"time": {"standard_step_size": 5}}})
    listener = Listener("listener")
    ## Define odds for validation, but don't pass it into the step modifier
    step_modifiers = {"evens": 3, "odds": 5}
    step_modifier_component = StepModifierWithRatePipeline(
        "step_modifier", step_modifiers["evens"], modified_simulants="evens"
    )
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    ## We should give odds standard size, and evens 3 days.
    correct_step_sizes = [3, 2, 1, 3, 1, 2, 3]
    groups = ["evens", "odds", "evens", "evens", "odds", "evens", "all"]

    ## Ensure that steps and active simulants are correct through one cycle of 15
    for correct_step_size, group in zip(correct_step_sizes, groups):
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=get_pop_by_parity(sim, group).index,
            expected_step_size_days=correct_step_size,
        )

        sample_pipeline = step_modifier_component.ts_pipeline_value
        assert sample_pipeline.index.equals(get_pop_by_parity(sim, group).index)
        assert np.all(sample_pipeline == pipeline_by_parity(sim, step_modifiers, group))


def test_multiple_modifiers(SimulationContext, base_config):
    """Test that if we have a mix of step sizes, we take steps in accordance
    to reach all simulants' next event times in the fewest steps.
    """

    listener = Listener("listener")
    step_modifiers = {"evens": 3, "odds": 7}
    step_modifier_A = StepModifier(
        "step_modifier_A", step_modifiers["evens"], modified_simulants="evens"
    )
    step_modifier_B = StepModifier(
        "step_modifier_B",
        step_modifiers["evens"],
        step_modifiers["odds"],
        modified_simulants="odds",
    )
    sim = SimulationContext(
        base_config,
        [step_modifier_A, step_modifier_B, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    ## With step sizes of 3 and 7, we need 3 steps, then 3 more, then one to get 7, then two more to get
    ## 9, etc.
    correct_step_sizes = [3, 3, 1, 2, 3, 2, 1, 3, 3]
    groups = ["evens", "evens", "odds", "evens", "evens", "odds", "evens", "evens", "all"]

    ## Ensure that steps and active simulants are correct through one cycle of 21
    for correct_step_size, group in zip(correct_step_sizes, groups):
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=get_pop_by_parity(sim, group).index,
            expected_step_size_days=correct_step_size,
        )


def test_untracked_simulants(SimulationContext, base_config):
    """Test that untracked simulants are always included in event indices, and are
    basically treated the same as any other simulant."""
    base_config.update({"configuration": {"time": {"standard_step_size": 7}}})
    listener = Listener("listener")
    step_modifier_component = StepModifierWithUntracking("step_modifier", 3)
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)

    for _ in range(2):
        take_step_and_validate(sim, listener, full_pop_index, expected_step_size_days=3)
        assert step_modifier_component.ts_pipeline_value.index.equals(full_pop_index)


def test_move_simulants_to_end(SimulationContext, base_config):
    """Ensure that we move simulants' next event time to the end of the simulation, if they are even."""
    base_config.update({"configuration": {"time": {"standard_step_size": 7}}})
    listener = Listener("listener")
    step_modifier_component = StepModifierWithMovement("step_modifier", 3)
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)
    odds = get_index_by_parity(full_pop_index, "odds")
    evens = get_index_by_parity(full_pop_index, "evens")
    take_step_and_validate(sim, listener, full_pop_index, expected_step_size_days=3)
    assert step_modifier_component.ts_pipeline_value.index.equals(full_pop_index)
    assert np.all(
        sim._clock.simulant_next_event_times(evens)
        == sim._clock.stop_time + sim._clock.minimum_step_size
    )

    for _ in range(2):
        take_step_and_validate(sim, listener, odds, expected_step_size_days=3)
        assert step_modifier_component.ts_pipeline_value.index.equals(odds)


def test_step_size_post_processor(builder):
    """Test that step size post-processor chooses the minimum modified step, or minimum global step,
    whichever is larger."""
    index = pd.Index(range(10))
    clock = SimulationClock()
    clock._standard_step_size = clock._minimum_step_size = pd.Timedelta(days=2)
    clock.setup(builder)

    ## Add modifier that set the step size to 7 for even indices and 5 for odd indices
    clock.register_step_modifier(
        lambda idx: pd.Series(
            [pd.Timedelta(days=7) if i % 2 == 0 else pd.Timedelta(days=5) for i in idx],
            index=idx,
        )
    )
    ## Add modifier that sets the step size to 9 for all simulants
    clock.register_step_modifier(lambda idx: pd.Series(pd.Timedelta(days=9), index=idx))
    value = clock._step_size_pipeline(index)
    evens = value.iloc[lambda x: x.index % 2 == 0]
    odds = value.iloc[lambda x: x.index % 2 == 1]

    ## The second modifier shouldn't have an effect
    assert np.all(evens == pd.Timedelta(days=6))
    assert np.all(odds == pd.Timedelta(days=4))

    clock.register_step_modifier(lambda idx: pd.Series(pd.Timedelta(days=0.5), index=idx))
    value = clock._step_size_pipeline(index)
    assert np.all(value == pd.Timedelta(days=2))


@pytest.mark.parametrize("end_day", [31, 23])
def test_time_steps_remaining(SimulationContext, base_config, end_day):
    UselessComponent = MockGenericComponent("Placeholder")
    base_config.update(
        {
            "configuration": {
                "time": {
                    "start": {"year": 2021, "month": 1, "day": 1},
                    "end": {"year": 2021, "month": 1, "day": end_day},
                    "step_size": 10,  # Days
                },
            },
        },
    )
    sim = SimulationContext(
        base_config,
        [UselessComponent],
    )
    sim.setup()
    sim.initialize_simulants()
    assert sim.get_number_of_steps_remaining() == 3
    sim.step()
    assert sim.get_number_of_steps_remaining() == 2


@pytest.mark.parametrize("end", [31, 23])
def test_simple_clock_time_steps_remaining(SimulationContext, base_config, end):
    UselessComponent = MockGenericComponent("Placeholder")
    base_config.update(
        {
            "configuration": {
                "time": {
                    "start": 1,
                    "end": end,
                    "step_size": 10,
                },
            },
            "plugins": {
                "required": {
                    "clock": {
                        "controller": "vivarium.framework.time.SimpleClock",
                    },
                }
            },
        },
    )
    sim = SimulationContext(
        base_config,
        [UselessComponent],
    )
    sim.setup()
    sim.initialize_simulants()
    assert sim.get_number_of_steps_remaining() == 3
    sim.step()
    assert sim.get_number_of_steps_remaining() == 2



import itertools
from typing import cast

import numpy as np
import pandas as pd
import pytest

from vivarium.framework.lookup.interpolation import (
    Interpolation,
    Order0Interp,
    check_data_complete,
    validate_parameters,
)


def make_bin_edges(data: pd.DataFrame, col: str) -> pd.DataFrame:
    """Given a dataframe and a column containing midpoints, construct
    equally sized bins around midpoints.
    """
    mid_pts = data[[col]].drop_duplicates().sort_values(by=col).reset_index(drop=True)
    mid_pts["shift"] = mid_pts[col].shift()

    mid_pts["left"] = mid_pts.apply(
        lambda row: (row[col] if pd.isna(row["shift"]) else 0.5 * (row[col] + row["shift"])),
        axis=1,
    )

    mid_pts["right"] = mid_pts["left"].shift(-1)
    mid_pts["right"] = mid_pts.right.fillna(
        mid_pts.right.max() + mid_pts.left.tolist()[-1] - mid_pts.left.tolist()[-2]
    )

    data = data.copy()
    idx = data.index

    data = data.set_index(col, drop=False)
    mid_pts = mid_pts.set_index(col, drop=False)

    data[[col, f"{col}_left", f"{col}_right"]] = mid_pts[[col, "left", "right"]]

    return data.set_index(idx)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_1d_interpolation() -> None:
    df = pd.DataFrame({"a": np.arange(100), "b": np.arange(100), "c": np.arange(100, 0, -1)})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, (), [("a", "a_start", "a_end")], ("c",), 1, True, True)

    query = pd.DataFrame({"a": np.arange(0, 100, step=0.01)})

    assert np.allclose(query.a, i(query).b)
    assert np.allclose(100 - query.a, i(query).c)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_age_year_interpolation() -> None:
    years = list(range(1990, 2010))
    ages = list(range(0, 90))
    pops = np.array(ages) * 11.1
    data = []
    for age, pop in zip(ages, pops):
        for year in years:
            for sex in ["Male", "Female"]:
                data.append({"age": age, "sex": sex, "year": year, "pop": pop})
    df = pd.DataFrame(data)

    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(
        df, ("sex", "age"), [("year", "year_start", "year_end")], ("pop",), 1, True, True
    )
    query = pd.DataFrame({"year": [1990, 1990], "age": [35, 35], "sex": ["Male", "Female"]})
    assert np.allclose(i(query), 388.5)


@pytest.mark.parametrize(
    "query",
    [
        pd.DataFrame({"year": [1990, 1990], "age": [35, 35]}),
        pd.DataFrame({"year": [1990, 1990], "sex": ["Male", "Female"]}),
    ],
)
@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_interpolation_called_missing_param_col(query: pd.DataFrame) -> None:
    a = [range(1990, 1995), range(25, 30), ["Male", "Female"]]
    df = pd.DataFrame(list(itertools.product(*a)), columns=["year", "age", "sex"])
    df["pop"] = df.age * 11.1
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input
    param_cols = [("year", "year_start", "year_end"), ("age", "age_start", "age_end")]
    i = Interpolation(df, ["sex"], param_cols, ["pop"], 1, True, True)
    with pytest.raises(ValueError):
        i(query)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_2d_interpolation() -> None:
    a = np.mgrid[0:5, 0:5][0].reshape(25)
    b = np.mgrid[0:5, 0:5][1].reshape(25)
    df = pd.DataFrame({"a": a, "b": b, "c": b, "d": a})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    param_cols = [("a", "a_start", "a_end"), ("b", "b_start", "b_end")]
    i = Interpolation(df, (), param_cols, ["c", "d"], 1, True, True)

    query = pd.DataFrame({"a": np.arange(0, 4, step=0.01), "b": np.arange(0, 4, step=0.01)})

    assert np.allclose(query.b, i(query).c)
    assert np.allclose(query.a, i(query).d)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_interpolation_with_categorical_parameters() -> None:
    a = ["one"] * 100 + ["two"] * 100
    b = np.append(np.arange(100), np.arange(100))
    c = np.append(np.arange(100), np.arange(100, 0, -1))
    df = pd.DataFrame({"a": a, "b": b, "c": c})
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(df, ("a",), [("b", "b_start", "b_end")], ["c"], 1, True, True)

    query_one = pd.DataFrame({"a": "one", "b": np.arange(0, 100, step=0.01)})
    query_two = pd.DataFrame({"a": "two", "b": np.arange(0, 100, step=0.01)})

    assert np.allclose(np.arange(0, 100, step=0.01), i(query_one).c)

    assert np.allclose(np.arange(0, 100, step=-0.01), i(query_two).c)


def test_order_zero_2d() -> None:
    a = np.mgrid[0:5, 0:5][0].reshape(25)
    b = np.mgrid[0:5, 0:5][1].reshape(25)
    df = pd.DataFrame({"a": a + 0.5, "b": b + 0.5, "c": b * 3, "garbage": ["test"] * len(a)})
    df = make_bin_edges(df, "a")
    df = make_bin_edges(df, "b")
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(
        df,
        ("garbage",),
        [("a", "a_left", "a_right"), ("b", "b_left", "b_right")],
        ["c"],
        order=0,
        extrapolate=True,
        validate=True,
    )

    column = np.arange(0.5, 4, step=0.011)
    query = pd.DataFrame({"a": column, "b": column, "garbage": ["test"] * (len(column))})

    assert np.allclose(query.b.astype(int) * 3, i(query).c)


def test_order_zero_2d_fails_on_extrapolation() -> None:
    a = np.mgrid[0:5, 0:5][0].reshape(25)
    b = np.mgrid[0:5, 0:5][1].reshape(25)
    df = pd.DataFrame({"a": a + 0.5, "b": b + 0.5, "c": b * 3, "garbage": ["test"] * len(a)})
    df = make_bin_edges(df, "a")
    df = make_bin_edges(df, "b")
    df = df.sample(frac=1)  # Shuffle table to assure interpolation works given unsorted input

    i = Interpolation(
        df,
        ("garbage",),
        [("a", "a_left", "a_right"), ("b", "b_left", "b_right")],
        ["c"],
        order=0,
        extrapolate=False,
        validate=True,
    )

    column = np.arange(0.0, 4.0, step=0.011)
    query = pd.DataFrame({"a": column, "b": column, "garbage": ["test"] * (len(column))})

    with pytest.raises(ValueError) as error:
        i(query)

    message = error.value.args[0]

    assert "Extrapolation" in message and "a" in message


def test_order_zero_1d_no_extrapolation() -> None:
    s = pd.Series({0: 0, 1: 1}, name="val").reset_index()
    s = make_bin_edges(s, "index")
    f = Interpolation(
        s,
        tuple(),
        [["index", "index_left", "index_right"]],
        ["val"],
        order=0,
        extrapolate=False,
        validate=True,
    )

    assert f(pd.DataFrame({"index": [0]}))["val"][0] == 0, "should be precise at index values"
    assert f(pd.DataFrame({"index": [0.999]}))["val"][0] == 1

    with pytest.raises(ValueError) as error:
        f(pd.DataFrame({"index": [1]}))

    message = error.value.args[0]
    assert "Extrapolation" in message and "index" in message


def test_order_zero_1d_constant_extrapolation() -> None:
    s = pd.Series({0: 0, 1: 1}, name="val").reset_index()
    s = make_bin_edges(s, "index")
    f = Interpolation(
        s,
        tuple(),
        [["index", "index_left", "index_right"]],
        ["val"],
        order=0,
        extrapolate=True,
        validate=True,
    )

    assert f(pd.DataFrame({"index": [1]}))["val"][0] == 1
    assert (
        f(pd.DataFrame({"index": [2]}))["val"][0] == 1
    ), "should be constant extrapolation outside of input range"
    assert f(pd.DataFrame({"index": [-1]}))["val"][0] == 0


def test_validate_parameters__empty_data() -> None:
    with pytest.raises(ValueError, match="must supply non-empty data"):
        validate_parameters(
            pd.DataFrame(
                columns=["age_left", "age_right", "sex", "year_left", "year_right", "value"]
            ),
            ["sex"],
            [("age", "age_left", "age_right"), ["year", "year_left", "year_right"]],
            ["value"],
        )


def test_check_data_complete_gaps() -> None:
    data = pd.DataFrame(
        {
            "year_start": [1990, 1990, 1995, 1995],
            "year_end": [1995, 1995, 2000, 2000],
            "age_start": [16, 10, 10, 16],
            "age_end": [20, 15, 15, 20],
        }
    )

    with pytest.raises(NotImplementedError) as error:
        check_data_complete(
            data, [("year", "year_start", "year_end"), ["age", "age_start", "age_end"]]
        )

    message = error.value.args[0]

    assert "age_start" in message and "age_end" in message


def test_check_data_complete_overlap() -> None:
    data = pd.DataFrame(
        {
            "year_start": [1995, 1995, 2000, 2005, 2010],
            "year_end": [2000, 2000, 2005, 2010, 2015],
        }
    )

    with pytest.raises(ValueError) as error:
        check_data_complete(data, [("year", "year_start", "year_end")])

    message = error.value.args[0]

    assert "year_start" in message and "year_end" in message


def test_check_data_missing_combos() -> None:
    data = pd.DataFrame(
        {
            "year_start": [1990, 1990, 1995],
            "year_end": [1995, 1995, 2000],
            "age_start": [10, 15, 10],
            "age_end": [15, 20, 15],
        }
    )

    with pytest.raises(ValueError) as error:
        check_data_complete(
            data, [["year", "year_start", "year_end"], ("age", "age_start", "age_end")]
        )

    message = error.value.args[0]

    assert "combination" in message


def test_order0interp() -> None:
    data = pd.DataFrame(
        {
            "year_start": [1990, 1990, 1990, 1990, 1995, 1995, 1995, 1995],
            "year_end": [1995, 1995, 1995, 1995, 2000, 2000, 2000, 2000],
            "age_start": [15, 10, 10, 15, 10, 10, 15, 15],
            "age_end": [20, 15, 15, 20, 15, 15, 20, 20],
            "height_start": [140, 160, 140, 160, 140, 160, 140, 160],
            "height_end": [160, 180, 160, 180, 160, 180, 160, 180],
            "value": [5, 3, 1, 7, 8, 6, 4, 2],
        }
    )

    interp = Order0Interp(
        data,
        [
            ("age", "age_start", "age_end"),
            ("year", "year_start", "year_end"),
            ("height", "height_start", "height_end"),
        ],
        ["value"],
        True,
        True,
    )

    interpolants = pd.DataFrame(
        {
            "age": [12, 17, 8, 24, 12],
            "year": [1992, 1998, 1985, 1992, 1992],
            "height": [160, 145, 140, 179, 160],
        }
    )

    result = interp(interpolants)
    assert result.equals(pd.DataFrame({"value": [3, 4, 1, 7, 3]}))


def test_order_zero_1d_with_key_column() -> None:
    data = pd.DataFrame(
        {
            "year_start": [1990, 1990, 1995, 1995],
            "year_end": [1995, 1995, 2000, 2000],
            "sex": ["Male", "Female", "Male", "Female"],
            "value_1": [10, 7, 2, 12],
            "value_2": [1200, 1350, 1476, 1046],
        }
    )

    param_cols = [("year", "year_start", "year_end")]
    i = Interpolation(data, ["sex"], param_cols, ["value_1", "value_2"], 0, True, True)

    query = pd.DataFrame(
        {
            "year": [
                1992,
                1993,
            ],
            "sex": ["Male", "Female"],
        }
    )

    expected_result = pd.DataFrame({"value_1": [10.0, 7.0], "value_2": [1200.0, 1350.0]})

    assert i(query).equals(expected_result)


def test_order_zero_non_numeric_values() -> None:
    data = pd.DataFrame(
        {
            "year_start": [1990, 1990],
            "year_end": [1995, 1995],
            "age_start": [
                15,
                24,
            ],
            "age_end": [24, 30],
            "value_1": ["blue", "red"],
        }
    )

    param_cols = [("year", "year_start", "year_end"), ("age", "age_start", "age_end")]
    i = Interpolation(data, tuple(), param_cols, ["value_1"], 0, True, True)

    query = pd.DataFrame(
        {
            "year": [1990, 1990],
            "age": [
                15,
                24,
            ],
        },
        index=[1, 0],
    )

    expected_result = pd.DataFrame({"value_1": ["blue", "red"]}, index=[1, 0])

    assert i(query).equals(expected_result)


def test_order_zero_3d_with_key_col() -> None:
    data = pd.DataFrame(
        {
            "year_start": [1990, 1990, 1990, 1990, 1995, 1995, 1995, 1995] * 2,
            "year_end": [1995, 1995, 1995, 1995, 2000, 2000, 2000, 2000] * 2,
            "age_start": [15, 10, 10, 15, 10, 10, 15, 15] * 2,
            "age_end": [20, 15, 15, 20, 15, 15, 20, 20] * 2,
            "height_start": [140, 160, 140, 160, 140, 160, 140, 160] * 2,
            "height_end": [160, 180, 160, 180, 160, 180, 160, 180] * 2,
            "sex": ["Male"] * 8 + ["Female"] * 8,
            "value": [5, 3, 1, 7, 8, 6, 4, 2, 6, 4, 2, 8, 9, 7, 5, 3],
        }
    )

    param_cols = [
        ("age", "age_start", "age_end"),
        ("year", "year_start", "year_end"),
        ("height", "height_start", "height_end"),
    ]
    interp = Interpolation(
        data,
        ("sex",),
        param_cols,
        ["value"],
        0,
        True,
        True,
    )

    interpolants = pd.DataFrame(
        {
            "age": [12, 17, 8, 24, 12],
            "year": [1992, 1998, 1985, 1992, 1992],
            "height": [160, 145, 140, 185, 160],
            "sex": ["Male", "Female", "Female", "Male", "Male"],
        },
        index=[10, 4, 7, 0, 9],
    )

    result = interp(interpolants)
    assert result.equals(
        pd.DataFrame({"value": [3.0, 5.0, 2.0, 7.0, 3.0]}, index=[10, 4, 7, 0, 9])
    )


def test_order_zero_diff_bin_sizes() -> None:
    data = pd.DataFrame(
        {
            "year_start": [
                1990,
                1995,
                1996,
                2005,
                2005.5,
            ],
            "year_end": [1995, 1996, 2005, 2005.5, 2010],
            "value": [1, 5, 2.3, 6, 100],
        }
    )

    param_cols = [("year", "year_start", "year_end")]
    i = Interpolation(data, tuple(), param_cols, ["value"], 0, False, True)

    query = pd.DataFrame({"year": [2007, 1990, 2005.4, 1994, 2004, 1995, 2002, 1995.5, 1996]})

    expected_result = pd.DataFrame({"value": [100, 1, 6, 1, 2.3, 5, 2.3, 5, 2.3]})

    assert i(query).equals(expected_result)


def test_order_zero_given_call_column() -> None:
    data = pd.DataFrame(
        {
            "year_start": [
                1990,
                1995,
                1996,
                2005,
                2005.5,
            ],
            "year_end": [1995, 1996, 2005, 2005.5, 2010],
            "year": [1992.5, 1995.5, 2000, 2005.25, 2007.75],
            "value": [1, 5, 2.3, 6, 100],
        }
    )

    param_cols = [("year", "year_start", "year_end")]
    i = Interpolation(data, tuple(), param_cols, ["value"], 0, False, True)

    query = pd.DataFrame({"year": [2007, 1990, 2005.4, 1994, 2004, 1995, 2002, 1995.5, 1996]})

    expected_result = pd.DataFrame({"value": [100, 1, 6, 1, 2.3, 5, 2.3, 5, 2.3]})

    assert i(query).equals(expected_result)


@pytest.mark.parametrize("validate", [True, False])
def test_interpolation_init_validate_option_invalid_data(validate: bool) -> None:
    if validate:
        with pytest.raises(
            ValueError, match="You must supply non-empty data to create the interpolation."
        ):
            i = Interpolation(pd.DataFrame(), [], [], [], 0, True, validate)
    else:
        i = Interpolation(pd.DataFrame(), [], [], [], 0, True, validate)


@pytest.mark.parametrize("validate", [True, False])
def test_interpolation_init_validate_option_valid_data(validate: bool) -> None:
    s = pd.Series({0: 0, 1: 1}, name="val").reset_index()
    s = make_bin_edges(s, "index")
    param_cols = [["index", "index_left", "index_right"]]
    Interpolation(s, tuple(), param_cols, ["val"], 0, True, validate)


@pytest.mark.parametrize("validate", [True, False])
def test_interpolation_call_validate_option_invalid_data(validate: bool) -> None:
    s = pd.Series({0: 0, 1: 1}, name="val").reset_index()
    s = make_bin_edges(s, "index")
    param_cols = [["index", "index_left", "index_right"]]
    i = Interpolation(s, tuple(), param_cols, ["val"], 0, True, validate)
    if validate:
        with pytest.raises(
            TypeError, match=r"Interpolations can only be called on pandas.DataFrames.*"
        ):
            result = i(cast(pd.DataFrame, 1))
    else:
        with pytest.raises(AttributeError):
            result = i(cast(pd.DataFrame, 1))


@pytest.mark.parametrize("validate", [True, False])
def test_interpolation_call_validate_option_valid_data(validate: bool) -> None:
    data = pd.DataFrame(
        {
            "year_start": [
                1990,
                1995,
                1996,
                2005,
                2005.5,
            ],
            "year_end": [1995, 1996, 2005, 2005.5, 2010],
            "value": [1, 5, 2.3, 6, 100],
        }
    )

    param_cols = [("year", "year_start", "year_end")]
    i = Interpolation(data, tuple(), param_cols, ["value"], 0, False, validate)
    query = pd.DataFrame({"year": [2007, 1990, 2005.4, 1994, 2004, 1995, 2002, 1995.5, 1996]})

    result = i(query)


@pytest.mark.parametrize("validate", [True, False])
def test_order0interp_validate_option_invalid_data(validate: bool) -> None:
    data = pd.DataFrame(
        {
            "year_start": [1995, 1995, 2000, 2005, 2010],
            "year_end": [2000, 2000, 2005, 2010, 2015],
        }
    )

    if validate:
        with pytest.raises(ValueError) as error:
            interp = Order0Interp(
                data, [("year", "year_start", "year_end")], [], True, validate
            )
            message = error.value.args[0]
            assert "year_start" in message and "year_end" in message
    else:
        interp = Order0Interp(data, [("year", "year_start", "year_end")], [], True, validate)


@pytest.mark.parametrize("validate", [True, False])
def test_order0interp_validate_option_valid_data(validate: bool) -> None:
    data = pd.DataFrame(
        {"year_start": [1990, 1995], "year_end": [1995, 2000], "value": [5, 3]}
    )

    interp = Order0Interp(
        data, [("year", "year_start", "year_end")], ["value"], True, validate
    )
